package org.netkernel.layer0.boot;

import java.util.*;

import org.netkernel.container.IKernel;
import org.netkernel.layer0.meta.IModuleMeta;
import org.netkernel.layer0.nkf.INKFLocale;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSIndex;
import org.netkernel.layer0.representation.IHDSIndexSpecification;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSFactory;
import org.netkernel.layer0.representation.impl.HDSIndex;
import org.netkernel.urii.ISpace;

public class BootOrderOptimizer
{
	public static BootOrderOptimizer sInstance=new BootOrderOptimizer();
	
	
	private List<ISpace> mStack=new ArrayList<ISpace>();
	
	private Map<ISpace,BootNode> mNodes = new IdentityHashMap<ISpace,BootNode>();
	private List<BootNode> mRoots = new ArrayList<BootNode>();
	private Map<IModule,Integer> mModules = Collections.EMPTY_MAP;
	private IHDSIndex mLoadedData;
	private Thread mThread;
	private int mStabilityThreshold;
	
	private static final boolean sEnabled=true;
	
	public boolean isActive()
	{	return mThread!=null;
	}
	
	
	public void analyze(List<IModule> aModules, INKFRequestContext aContext)
	{
		mThread=null;
		
		/*
		//find spaces which are initialised the most
		List<BootNode> nodes = new ArrayList<BootNode>(mNodes.size());
		nodes.addAll(mNodes.values());
		Comparator<BootNode> c = new Comparator<BootNode>()
		{
			public int compare(BootNode o1, BootNode o2)
			{
				return o1.mCount-o2.mCount;
			}
		};
		Collections.sort(nodes, c);
		for (BootNode node: nodes)
		{	System.out.println(node.mCount+" "+node.mSpace);
		}
		*/
		
		BootNode mostUsed=null;
		for (BootNode node : mNodes.values())
		{
			if (mostUsed==null || mostUsed.mCount<node.mCount)
			{	mostUsed=node;
			}
		}
		if (mostUsed!=null && mostUsed.mCount>mStabilityThreshold)
		{	aContext.logRaw(INKFLocale.LEVEL_WARNING, "Possible startup instability ["+mostUsed.mCount+"] with ["+mostUsed.mSpace+"] - consider increasing stdmod.initialisation.hysteresis in kernel.properties");
		}
		
		if (sEnabled)
		{
			for (BootNode node: mRoots)
			{	List<BootNode> l=new ArrayList<BootNode>();
				recurseNode1(node,0,l);
			}
			
			//find spaces which are initialised the most
			List<BootNode> nodes = new ArrayList<BootNode>(mNodes.size());
			nodes.addAll(mNodes.values());
			Comparator<BootNode> c = new Comparator<BootNode>()
			{
				public int compare(BootNode o1, BootNode o2)
				{
					return o1.mDepth-o2.mDepth;
				}
			};
			Collections.sort(nodes, c);
			for (BootNode node: nodes)
			{	//System.out.println("DEPTH: "+node.mDepth+" "+node.mSpace);
			}
			
			
			
			Map<IModule,Integer> modules=new HashMap<IModule, Integer>();
			for (IModule module : aModules)
			{
				ISpace[] spaces=module.getSpaces();
				int d=0;
				for (ISpace space : spaces)
				{
					int d2;
					BootNode node=mNodes.get(space);
					if (node!=null)
					{	d2=node.mDepth;
					}
					else
					{	d2=0;
					}
					if (d2>d) d=d2;
				}
				modules.put(module, Integer.valueOf(d));			
			}
			
			mModules=modules;
		}
			
	}
	
	private static int recurseNode1(BootNode aNode, int aDepth, List<BootNode> aList)
	{
		if (aList.contains(aNode))
		{
			/*
			StringBuffer sb=new StringBuffer();
			sb.append("circularity: ");
			boolean enabled=false;
			for (BootNode node: aList)
			{	if (node==aNode) enabled=true;
				if (enabled)
				{
					sb.append(node.mSpace.toString());
					sb.append(" --> ");
				}
			}
			sb.append(aNode.mSpace.toString());
			System.out.println(sb.toString());
			*/
			return Integer.MAX_VALUE;
		}
		aList.add(aNode);
		int d=0;
		for (BootNode node: aNode.mChildren)
		{
			int d2=recurseNode1(node,aDepth+1,aList);
			if (d2>d)
			{	d=d2;
			}
		}
		aNode.mDepth=d;
		aList.remove(aList.size()-1);
		return d+1;
	}
	
	public void reset(IKernel aKernel)
	{
		mThread=Thread.currentThread();
		mRoots.clear();
		mNodes.clear();		
		mStabilityThreshold=aKernel.getConfiguration().getInt("netkernel.init.stabilityThreshold", 10);
	}
	
	
	public void store(INKFRequestContext aContext) throws Exception
	{	if (!BootUtils.bootedFromJar(aContext.getKernelContext().getKernel().getConfiguration()))
		{	HDSBuilder b=new HDSBuilder();
			b.pushNode("boo");
			for (Map.Entry<IModule, Integer> entry : mModules.entrySet())
			{
				b.pushNode("module");
				IModuleMeta meta=entry.getKey().getMeta();
				b.addNode("uri",meta.getIdentifier());
				b.addNode("version",meta.getVersion());
				b.addNode("order",entry.getValue());
				b.popNode();
			}
			IBinaryStreamRepresentation bs=aContext.transrept(b.getRoot(), IBinaryStreamRepresentation.class);
			BootPersistence.sinkBootResource("boo.xml", bs, aContext.getKernelContext().getKernel());
		}
	}
	
	
	
	public void init(IKernel aKernel)
	{
		if (mLoadedData==null)
		{
			try
			{	IReadableBinaryStreamRepresentation rep=BootPersistence.sourceBootResource("boo.xml", aKernel);
				IHDSNode data=HDSFactory.parseXML(rep.getInputStream());
				
				mLoadedData=new HDSIndex(data, sIndexSpec);
			}
			catch (Exception e)
			{ //dummy - no data persisted
				HDSBuilder b=new HDSBuilder();
				b.pushNode("boo");
				mLoadedData=new HDSIndex(b.getRoot(), sIndexSpec);
			}
		}
		
	}
	
	private static final IHDSIndexSpecification sIndexSpec = new IHDSIndexSpecification()
	{
		public Object getKeyValue(IHDSNode aNodeToIndex)
		{	return aNodeToIndex.getFirstValue("uri");
		}

		public IHDSNodeList getNodesToIndex(IHDSNode aContextNode)
		{	return aContextNode.getNodes("/boo/module");
		}
	};
	
	public void sort(List<IModule> aModules)
	{
		Comparator<IModule> c = new Comparator<IModule>()
		{
			public int compare(IModule o1, IModule o2)
			{	
				int d1=getModuleSortOrder(o1);
				int d2=getModuleSortOrder(o2);
				return d1-d2;
			}
		};
		if (sEnabled) Collections.sort(aModules,c);
	}
	
	
	private int getModuleSortOrder(IModule aModule)
	{
		Integer d=mModules.get(aModule);
		if (d==null)
		{	
			String uri=aModule.getMeta().getIdentifier();
			IHDSNode loadedData=mLoadedData.get(uri).getFirstNode();
			if (loadedData!=null)
			{	d=(Integer)loadedData.getFirstValue("order");
			}
			else
			{	d=0;
			}
		}
		return d;
	}
	
	
	
	public void enterSpace(ISpace aSpace, boolean aFromCache)
	{	
		if (mThread==Thread.currentThread())
		{
			/*
			int d=mStack.size();
			
			String st="                                                                                                   ".substring(0,d*2+1);
			if (mStack.size()==0)
			{	st="*";
			}
			//String t=Thread.currentThread().toString();
			System.out.println(st+aSpace);
			*/
			
			int s=mStack.size();
			BootNode node=mNodes.get(aSpace);
			if (node==null)
			{	node=new BootNode(aSpace);
				mNodes.put(aSpace, node);
				
				if (s==0)
				{	mRoots.add(node);
				}
				
			}
			if (!aFromCache)
			{	int count=node.touch();
				if (count>mStabilityThreshold)
				{	//System.out.println("STABILITY: "+count+" "+aSpace);
					//try { Thread.sleep(60000); } catch (Exception e) {;}
				}
			}
			
			if (s>0)
			{	ISpace parentSpace=mStack.get(s-1);
				if (mStack.contains(aSpace))
				{
					//System.out.println("rentry: "+aSpace);
				}
				else
				{	BootNode parentNode=mNodes.get(parentSpace);
					if (!parentNode.mChildren.contains(node))
					{	parentNode.addChild(node);
						mRoots.remove(node);
					}
				}
			}
			mStack.add(aSpace);
		}
	}
	
	public void exitSpace(ISpace aSpace)
	{
		if (mThread==Thread.currentThread())
		{	
			if (mStack.size()==0)
			{	//System.out.println("!!!!!!!!!BOO problem detected");
			}
			else
			{
				ISpace sp=mStack.remove(mStack.size()-1);
				if (sp!=aSpace)
				{	System.out.println("hmm - expected "+sp+" but got "+aSpace);
				}
			}
		}
	}
	
	
	private static class BootNode
	{
		public final ISpace mSpace;
		public final List<BootNode> mChildren=new ArrayList<BootNode>();
		public int mDepth;
		public int mCount;
		
		public BootNode(ISpace aSpace)
		{	mSpace=aSpace;
		}
		
		public void addChild(BootNode aNode)
		{	mChildren.add(aNode);
		}
		
		public int touch()
		{	return mCount++;
		}
	}
	
}
