package org.netkernel.layer0.urii;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.netkernel.container.IMessages;
import org.netkernel.layer0.bnf.BNFGrammarFactory;
import org.netkernel.layer0.bnf.ExtraneousPartsException;
import org.netkernel.layer0.meta.GrammarException;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.util.OrderedCheapMultiStringMap;
import org.netkernel.layer0.util.PairList;
import org.netkernel.urii.IIdentifier;

public class ParsedIdentifierImpl implements IIdentifier
{
	private String mToString;
	private String[] mArguments;
	private IIdentifierGrammar mGrammar;
	
	public static final String ARG_ACTIVE_SCHEME="scheme";
	public static final String ARG_ACTIVE_TYPE="activeType";
	public static final String ARG_ACTIVE_FRAGMENT="fragment";
	
	private ParsedIdentifierImpl(String aToString, String[] aArguments, IIdentifierGrammar aGrammar)
	{	mToString=aToString;
		mArguments=aArguments;
		if (aArguments.length%2==1)
		{	throw new IllegalArgumentException("aArguments");
		}
		mGrammar=aGrammar;
	}
	
	public static ParsedIdentifierImpl parseActiveSyntax(String aBase, OrderedCheapMultiStringMap aArgs)
	{
		OrderedCheapMultiStringMap args = new OrderedCheapMultiStringMap(aArgs);
		args.sort(); // sorting is important for stable identifier construction
		
		StringBuffer sb=new StringBuffer(256);
		sb.append(aBase);
		for(int i=0; i<args.size(); i++)
		{	sb.append('+');
			sb.append(args.getKey(i));
			sb.append('@');
			sb.append(ActiveURIEncoding.INSTANCE.encode(args.getValue(i)));
		}
		
		int i=aBase.indexOf(':');
		String scheme;
		String type;
		if (i>=0)
		{	scheme=aBase.substring(0,i);
			type=aBase.substring(i+1);
		}
		else
		{	scheme="";
			type=aBase;
		}
		args.put(ARG_ACTIVE_SCHEME, scheme);
		args.put(ARG_ACTIVE_TYPE, type);
		
		return new ParsedIdentifierImpl(sb.toString(),args.asArray(),null);
	}
	
	public static ParsedIdentifierImpl parseActiveSyntax(String aIdentifier)
	{
		int iColon=aIdentifier.indexOf(':');
		String scheme=aIdentifier.substring(0,iColon);
		String fragment;
		int iHash=aIdentifier.lastIndexOf('#');
		if (iHash<0)
		{	iHash=aIdentifier.length();
			fragment=null;
		}
		else
		{	fragment=aIdentifier.substring(iHash+1);
		}
		
		String content=aIdentifier.substring(iColon+1,iHash);
		String type=null;
		int length = content.length();
		PairList args;
		int i=0;
		int iPlus = content.indexOf('+',i);
		if (iPlus>0)
		{	type=content.substring(0,iPlus);
			i=iPlus+1;
			args = new PairList(8);
			while(i<length)
			{	iPlus = content.indexOf('+',i);
				if (iPlus<0) iPlus=length;
				int iAt = content.indexOf('@',i);
				if (iAt>0 && iPlus>iAt)
				{	String key = content.substring(i,iAt);
					String uri = ActiveURIEncoding.INSTANCE.decode(content.substring(iAt+1,iPlus));
					args.put(key,uri);
				}
				else if (iAt>0 || iPlus>0)
				{	args=null;
					break;
				}
				i=iPlus+1;
			}
		}
		else
		{	args=null;
		}
		
		int argSize=((args!=null)?args.size():0) + ((fragment!=null)?1:0) + 2;
		String[] aa = new String[argSize*2];
		aa[0]=ARG_ACTIVE_SCHEME;
		aa[1]=scheme;
		aa[2]=ARG_ACTIVE_TYPE;
		aa[3]=type;
		int j=4;
		if (args!=null)
		{	for (i=0; i<args.size(); i++)
			{	aa[j++]=(String)args.getValue1(i);
				aa[j++]=(String)args.getValue2(i);
			}
		}
		if (fragment!=null)
		{	aa[j++]=ARG_ACTIVE_FRAGMENT;
			aa[j++]=fragment;
		}
		
		return new ParsedIdentifierImpl(aIdentifier,aa,null);		
	}
	
	public static ParsedIdentifierImpl parseGrammar(String aIdentifier, IIdentifierGrammar aGrammar) throws GrammarException
	{
		String[] arguments=aGrammar.parse(aIdentifier).asArray();
		
		//replace argName/argValue pairs
		String[] arg2 = new String[arguments.length*2];
		int j=0;
		int l2=arguments.length-2;
		for (int i=0; i<arguments.length; i+=2)
		{	if (i<l2 && arguments[i].equals(BNFGrammarFactory.ARG_UNDEFINED_NAME) && arguments[i+2].equals(BNFGrammarFactory.ARG_UNDEFINED_VALUE))
			{	String argName=arguments[i+1];
				String argValue=arguments[i+3];
				arg2[j++]=argName;
				arg2[j++]=argValue;
				i+=2;
			}
			else
			{	arg2[j++]=arguments[i];
				arg2[j++]=arguments[i+1];
			}
		}
		
		if (j!=arguments.length)
		{	arguments = new String[j];
			System.arraycopy(arg2, 0, arguments, 0, j);
		}
		
		return new ParsedIdentifierImpl(aIdentifier,arguments,aGrammar);
	}
	
	public static ParsedIdentifierImpl parseGrammar(OrderedCheapMultiStringMap aArguments, IIdentifierGrammar aGrammar, IMessages aMessages) throws GrammarException
	{	try
		{	String identifier=aGrammar.construct(aArguments,aMessages);
			OrderedCheapMultiStringMap args=aGrammar.parse(identifier);
			ParsedIdentifierImpl result = new ParsedIdentifierImpl(identifier,args.asArray(),aGrammar);
			return result;
		}
		catch (ExtraneousPartsException e)
		{	// extra parts, try converting remaining parts to varargs
			OrderedCheapMultiStringMap args2 = new OrderedCheapMultiStringMap(aArguments);
			OrderedCheapMultiStringMap toRemove = e.getExtraParts();
			for (int i=0; i<toRemove.size(); i++)
			{	String name=toRemove.getKey(i);
				String value=toRemove.getValue(i);
				args2.remove(name);
				args2.put(BNFGrammarFactory.ARG_UNDEFINED_NAME, name);
				args2.put(BNFGrammarFactory.ARG_UNDEFINED_VALUE, value);
			}
			try
			{	String identifier=aGrammar.construct(args2,aMessages);
				ParsedIdentifierImpl result = new ParsedIdentifierImpl(identifier,aArguments.asArray(),aGrammar);
				return result;
			}
			catch (GrammarException e2)
			{	// throw the original exception if this fails
				throw e;
			}
		}
	}

	public int getArgumentCount()
	{	return mArguments.length>>1;
	}
	
	public String getArgumentKey(int aIndex)
	{	return mArguments[aIndex<<1];
	}
	
	public String getArgumentValue(int aIndex)
	{	return mArguments[1+(aIndex<<1)];
	}
	
	public String getArgumentValue(String aKey)
	{	String result=null;
		for (int i=mArguments.length-2; i>=0; i-=2)
		{	if (mArguments[i].equals(aKey))
			{	result=mArguments[i+1];
				break;
			}
		}
		return result;
	}
	
	public List<String> getArgumentValues(String aKey)
	{
		List<String> result=new ArrayList<String>(4);
		for (int i=mArguments.length-2; i>=0; i-=2)
		{	if (mArguments[i].equals(aKey))
			{	result.add(mArguments[i+1]);
			}
		}
		if (result.size()==0)
		{	result=Collections.EMPTY_LIST;
		}
		return result;
	}
	
	public String toString()
	{	return mToString;
	}
	
	public boolean equals(Object aOther)
	{	return mToString.equals(aOther.toString());
	}
	
	public int hashCode()
	{	return mToString.hashCode();
	}	
	
	/** Returns null if no grammar is associated - i.e. maybe active identifier
	 */
	public IIdentifierGrammar getGrammar()
	{	return mGrammar;
	}
}