/******************************************************************************
 * (c) Copyright 2002,2003, 1060 Research Ltd
 *
 * This Software is licensed to You, the licensee, for use under the terms of
 * the 1060 Public License v1.0. Please read and agree to the 1060 Public
 * License v1.0 [www.1060research.com/license] before using or redistributing
 * this software.
 *
 * In summary the 1060 Public license has the following conditions.
 * A. You may use the Software free of charge provided you agree to the terms
 * laid out in the 1060 Public License v1.0
 * B. You are only permitted to use the Software with components or applications
 * that provide you with OSI Certified Open Source Code [www.opensource.org], or
 * for which licensing has been approved by 1060 Research Limited.
 * You may write your own software for execution by this Software provided any
 * distribution of your software with this Software complies with terms set out
 * in section 2 of the 1060 Public License v1.0
 * C. You may redistribute the Software provided you comply with the terms of
 * the 1060 Public License v1.0 and that no warranty is implied or given.
 * D. If you find you are unable to comply with this license you may seek to
 * obtain an alternative license from 1060 Research Limited by contacting
 * license@1060research.com or by visiting www.1060research.com
 *
 * NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 * LICENSE V1.0 FOR DETAILS
 *
 * THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 * THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license
 *
 * File:          $RCSfile$
 * Version:       $Name$ $Revision$
 * Last Modified: $Date$
 *****************************************************************************/
package org.netkernel.layer0.util;


import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

import org.netkernel.layer0.representation.IHDSNode;
import org.w3c.dom.Node;

/**
 * A Fast implementation XPath for simple expressions.
 * Evaluation shows it is an order of magnitude faster than xalan and will
 * work for about 2/3 of expressions in normal usage.
 * <br/>It will evaluate xpaths of the forms:
 * <ul><li> /a/b (absolute)
 * </li><li> a/b (relative)
 * </li><li> /a/* /c (disregard name)
 * </li><li> /a/b[1] (element position)
 * </li><li> /a/@b (attribute)
 * </li><li> //a (any element a below cursor)
 * </li></ul>
 * @author  tab
 */
public class HDSXPath
{	/** all expressions are cached for quick evaluation */
	private static Map mCachedEvals = new ConcurrentHashMap(128);

	/** Inner class to store pre-parsed xpath expression */
	private static class SimpleEvalStruct
	{	String[] mNames;
		int[] mIndices;
		
		private static final int INDEX_ANY=-1;
		private static final int INDEX_ANYDEPTH=-2;
		
		/** construct/parse a new xpath expression */
		public SimpleEvalStruct(String aXPath)
		{	
			StringTokenizer stz = new StringTokenizer(aXPath,"/",true);
			int length = stz.countTokens();
			String[]  names = new String[length];
			int[] indices = new int[length];
			int c=0;
			int lastSlash=-2;
			boolean anydepth=false;
			for (int i=0; i<length; i++)
			{	String token = stz.nextToken();
				
				if (!token.equals("/"))
				{	
					//System.out.println("token of ["+token+"]");
					int i1 = token.indexOf('[');
					if (i1>=0)
					{	int i2 = token.indexOf(']');
						String indexString = token.substring(i1+1,i2);
						try
						{	indices[c] = Integer.parseInt(indexString);
						} catch (NumberFormatException e)
						{	throw new IllegalArgumentException(aXPath+": Predicates limited to integer expressions");
						}
						names[c] = token.substring(0,i1);
					}
					else if (anydepth)
					{	indices[c]=INDEX_ANYDEPTH;
						names[c] = token;
					}
					else
					{	indices[c]=INDEX_ANY;
						names[c] = token;
					}
					if (names[c].equals("*"))
					{	names[c]=null;
					}
					c++;
					anydepth=false;
				}
				else
				{	if (lastSlash==i-1)
					{	//anydepth slash slash
						anydepth=true;
					}
					lastSlash=i;
					
				}
				
			}
			mNames = new String[c];
			System.arraycopy(names,0,mNames,0,c);
			mIndices = new int[c];
			System.arraycopy(indices,0,mIndices,0,c);
			
		}
		
		public void eval(List<IHDSNode> aResult, IHDSNode aContextNode, boolean aSingleMatch)
		{	if (size()>0)
			{	evalDescend(aResult,aContextNode,0,aSingleMatch);
			}
			else
			{	aResult.add(aContextNode);
			}
		}
		
		/** internal method to perform the iteration down the document
		 * @param aResult the nodeset to add final results too
		 * @param aCurrent the current node
		 * @param aDepth the current depth
		 */
		private boolean evalDescend(List<IHDSNode> aResult, IHDSNode aCurrent, int aDepth, boolean aSingleMatch)
		{	boolean result=false;
			if (aDepth!=size()-1)
			{	List<IHDSNode> ns = new ArrayList<IHDSNode>(aCurrent.getChildren().length);
				findMatch(ns,aCurrent,aDepth);
				int upperLimit = ns.size();  // Ask for the size just once
				for(int i=0; i<upperLimit; i++)
				{	IHDSNode n = ns.get(i);
					if (evalDescend(aResult,n,aDepth+1,aSingleMatch) && aSingleMatch)
					{	break;
					}
				}
			}
			else
			{	findMatch(aResult,aCurrent,aDepth);
				result=true;
			}
			return result;
		}			
		
		
		/** finds all nodes that match as the children of the given node at the
		 * given depth 
		 */
		private void findMatch(List<IHDSNode> aResult, IHDSNode aNode, int aDepth)
		{	
			int foundIndex=0;
			String name = mNames[aDepth];
			int index = mIndices[aDepth];
			
			if (index==INDEX_ANYDEPTH)
			{	//anydepth search
				IHDSNode n=aNode;
				while (n!=null)
				{	if (name==null || name.equals(n.getName()))
					{	aResult.add(n);
					}
					n=HDSUtils.inOrderTraversalNext(n,aNode);
				}
			}
			else
			{	IHDSNode[] nodes=aNode.getChildren();
				for (int i=0; i<nodes.length; i++)
				{	IHDSNode n=nodes[i];
					if (name==null || name.equals(n.getName()))
					{	foundIndex++;
						if (index==INDEX_ANY || foundIndex==index)
						{	aResult.add(n);
						}
					}
				}
			}
		}
		
		
		
		
		
		
		/** return the depth of the xpath expression */
		public final int size()
		{	return mNames.length;
		}
	}
	
	
	
	
	
	
	
	/** Returns a nodeset of matching nodes for the evaluation of the xpath
	 * at the given context node
	 */
	public static void eval(List<IHDSNode> aResults, IHDSNode aContextNode, String aXPath)
	{	innerEval(aResults, aContextNode,aXPath,false);
	}
	
	public static IHDSNode getSingleNode(IHDSNode aContextNode, String aXPath, boolean aFailOnError) throws NoSuchElementException
	{	List<IHDSNode> ns = new ArrayList<IHDSNode>();
		innerEval(ns,aContextNode,aXPath,true);
		if (ns.size()==0)
		{	if (aFailOnError)
			{	throw new NoSuchElementException(aXPath);
			}
			else
			{	return null;
			}
		}
		return (IHDSNode)ns.get(0);
	}

	
	private static void innerEval(List<IHDSNode> aResults, IHDSNode aContextNode, String aXPath, boolean aSingleMatch)
	{		
		// parse simple xpath and cache
		SimpleEvalStruct levels = (SimpleEvalStruct)mCachedEvals.get(aXPath);
		if (levels==null)
		{	levels = new SimpleEvalStruct(aXPath);
			mCachedEvals.put(aXPath,levels);
		}
		
		if (aXPath.charAt(0)=='/' && levels.size()>0 && levels.mIndices[0]!=SimpleEvalStruct.INDEX_ANYDEPTH)
		{	while (aContextNode.getParent()!=null)
			{	aContextNode = aContextNode.getParent();
			}
		}	
		
		levels.eval(aResults, aContextNode,aSingleMatch);
	}

}