package org.netkernel.layer0.util;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.*;

import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.request.IRequestScopeLevel;

/** Provides a mechanism for SOURCEing resources from a NetKernel context via a java.net.URLConnection. This
 * is useful for technologies that only know about URLs and cannot issue NetKernel requests.
 * @author tab
 *
 */
public class NKProtocolHandlerRegistry
{
	private static class ContextScopePair
	{	private final INKFRequestContext mContext;
		private final IRequestScopeLevel mScope;
		public ContextScopePair(INKFRequestContext aContext, IRequestScopeLevel aScope)
		{	mContext=aContext;
			mScope=aScope;
		}
		public INKFRequestContext getContext()
		{	return mContext;
		}
		public IRequestScopeLevel getScope()
		{	return mScope;
		}
		public int hashCode()
		{	return mContext.hashCode()%mScope.hashCode();
		}
		public boolean equals(Object aOther)
		{	boolean result=aOther instanceof ContextScopePair;
			if (result)
			{	ContextScopePair other = (ContextScopePair)aOther;
				result = mContext==other.mContext && mScope==other.mScope;
			}
			return result;
		}
	}
	private static final Map<String,ContextScopePair> sUUIDToPair = new HashMap<String, ContextScopePair>();
	private static final Map<ContextScopePair,String> sPairToUUID = new HashMap<ContextScopePair, String>();
	private static int sUUID = 0;
	
	/** This method is called by the registered nk: scheme prototcol handler
	 * @param aURL the resource URL
	 * @return an inputstream for a representation of the resource
	 * @throws Exception if the resource cannot be SOURCED as a readable binary stream.
	 */
	public static InputStream getInputStream(URL aURL) throws Exception
	{
		String path= aURL.getPath();
	    int i= path.indexOf(":");
	    String uuid= path.substring(0, i);
	    path= path.substring(i + 1);
	    
	    ContextScopePair pair = sUUIDToPair.get(uuid);
	    if (pair==null)
	    {	throw new IOException("URL "+aURL.getPath()+" cannot be resolved");
	    }
	    INKFRequestContext c=pair.getContext();
	    
	    INKFRequest req = c.createRequest(path);
	    req.setRequestScope(pair.getScope());
	    req.setRepresentationClass(IReadableBinaryStreamRepresentation.class);
	    IReadableBinaryStreamRepresentation rbs = (IReadableBinaryStreamRepresentation)c.issueRequest(req);
	    return rbs.getInputStream();
	}
	
	/** Generate a global URL for the given local identifier in the given NetKernel context.
	 * @param aContext the context in which the identifier should be resolved
	 * @param aLocalIdentifier the identifier to SOURCE
	 * @return an identifier that can be used to construct a java.net.URL that will
	 * be resolvable with the URLConnection infrastructure. 
	 */
	public static String getGlobalURL(INKFRequestContext aContext, String aLocalIdentifier)
	{	return getGlobalURL(aContext,aContext.getKernelContext().getRequestScope(),aLocalIdentifier);
	}

	/** Generate a global URL for the given local identifier in the given NetKernel context with the given request scope.
	 * @param aContext the context in which the identifier should be resolved
	 * @param aScope the request scope to use to resolve the request.
	 * @param aLocalIdentifier the identifier to SOURCE
	 * @return an identifier that can be used to construct a java.net.URL that will
	 * be resolvable with the URLConnection infrastructure. 
	 */	
	public static String getGlobalURL(INKFRequestContext aContext, IRequestScopeLevel aScope, String aLocalIdentifier)
	{	ContextScopePair pair = new ContextScopePair(aContext,aScope);
		String uuid;
		synchronized(sUUIDToPair)
		{	uuid = sPairToUUID.get(pair);
			if (uuid==null)
			{	uuid = Integer.toString(sUUID++);
				sPairToUUID.put(pair,uuid);
				sUUIDToPair.put(uuid,pair);
			}
		}
		StringBuilder sb = new StringBuilder(aLocalIdentifier.length()+uuid.length()+4);
		sb.append("nk:");
		sb.append(uuid);
		sb.append(":");
		sb.append(aLocalIdentifier);
		return sb.toString();
	}
	
	/** Remove mapping of a Context and scope to global URLs and release memory.
	 * @param aContext the context used when originally getting a global URL
	 */
	public void deregister(INKFRequestContext aContext)
	{	deregister(aContext,aContext.getKernelContext().getRequestScope());
	}	
	
	/** Remove mapping of a Context and scope to global URLs and release memory.
	 * @param aContext the context used when originally getting a global URL
	 * @param aScope the scope used when originally getting a global URL
	 */
	public void deregister(INKFRequestContext aContext, IRequestScopeLevel aScope)
	{	ContextScopePair pair = new ContextScopePair(aContext,aScope);
		synchronized(sUUIDToPair)
		{	String uuid=sPairToUUID.remove(sUUIDToPair);
			if (uuid!=null)
			{	sUUIDToPair.remove(uuid);
			}
		}
	}
	
}
